/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s):
 *   Peter Hartshorn <peter@igelaus.com.au>
 *   B.J. Rossiter <bj@igelaus.com.au>
 */

#include "nsDrawingSurfaceNanoX.h"

#include <nano-X.h>
#include <stdlib.h>

static NS_DEFINE_IID(kIDrawingSurfaceIID, NS_IDRAWING_SURFACE_IID);

nsDrawingSurfaceNanoX::nsDrawingSurfaceNanoX()
{
  NS_INIT_REFCNT();

  mDrawable = 0;
  mDepth = 0;
  mGC = 0;

  mLocked = PR_FALSE;
  mLockX = 0;
  mLockY = 0;
  mLockWidth = 0;
  mLockHeight = 0;
  mLockFlags = 0;
  pixels = 0;

  mWidth = 0;
  mHeight = 0;
  mIsOffscreen = PR_FALSE;

  mPixFormat.mRedMask = 0xff0000;
  mPixFormat.mGreenMask = 0x00ff00;
  mPixFormat.mBlueMask = 0x0000ff;
  mPixFormat.mAlphaMask = 0x0;

  mPixFormat.mRedCount = ConvertMaskToCount(mPixFormat.mRedMask);
  mPixFormat.mGreenCount = ConvertMaskToCount(mPixFormat.mGreenMask);
  mPixFormat.mBlueCount = ConvertMaskToCount(mPixFormat.mBlueMask);

  mPixFormat.mRedShift = GetShiftForMask(mPixFormat.mRedMask);
  mPixFormat.mGreenShift = GetShiftForMask(mPixFormat.mGreenMask);
  mPixFormat.mBlueShift = GetShiftForMask(mPixFormat.mBlueMask);

  mRegion = 0;
}

nsDrawingSurfaceNanoX::~nsDrawingSurfaceNanoX()
{
  if ((mDrawable) && (mOurDrawable))
    GrDestroyWindow(mDrawable);

  if (mGC)
    GrDestroyGC(mGC);

  if (mRegion)
    GrDestroyRegion(mRegion);
}

NS_IMPL_QUERY_INTERFACE(nsDrawingSurfaceNanoX, kIDrawingSurfaceIID)
NS_IMPL_ADDREF(nsDrawingSurfaceNanoX)
NS_IMPL_RELEASE(nsDrawingSurfaceNanoX)

NS_IMETHODIMP nsDrawingSurfaceNanoX::Init(GR_WINDOW_ID aDrawable)
{
  mDrawable = aDrawable;
  mOurDrawable = PR_FALSE;
  mGC = GrNewGC();
  GrSetGCUseBackground(mGC, PR_FALSE);
  mIsOffscreen = PR_FALSE;
  return NS_OK;
}

NS_IMETHODIMP nsDrawingSurfaceNanoX::Init(PRUint32 aWidth,
                                          PRUint32 aHeight,
                                          PRUint32 aFlags)
{
  if (aWidth <= 0) aWidth = 1;
  if (aHeight <= 0) aHeight = 1;
  mWidth = aWidth;
  mHeight = aHeight;
  mLockFlags = aFlags;
  mIsOffscreen = PR_TRUE;

  mDrawable = GrNewPixmap(mWidth, mHeight, nsnull);
  mOurDrawable = PR_TRUE;

  mGC = GrNewGC();
  GrSetGCUseBackground(mGC, PR_FALSE);

  return NS_OK;
}

NS_IMETHODIMP nsDrawingSurfaceNanoX::Lock(PRInt32 aX, PRInt32 aY,
                                          PRUint32 aWidth, PRUint32 aHeight,
                                          void **aBits, PRInt32 *aStride,
                                          PRInt32 *aWidthBytes, PRUint32 aFlags)
{
  if (mLocked) {
    NS_ASSERTION(0, "nested lock attempt");
    return NS_ERROR_FAILURE;
  }
  if (aWidth == 0 || aHeight == 0) {
    NS_ASSERTION(0, "Width or height is 0");
    return NS_ERROR_FAILURE;
  }

  mLocked = PR_TRUE;

  mLockX = aX;
  mLockY = aY;
  mLockWidth = aWidth;
  mLockHeight = aHeight;
  if (!pixels)
    pixels = (GR_PIXELVAL*)malloc(mLockWidth * mLockHeight * 4);

  GrReadArea(mDrawable, mLockX, mLockY, mLockWidth, mLockHeight, pixels);

  *aBits = pixels;
  *aWidthBytes = mLockWidth * 4;
  *aStride = mLockWidth * 4;

  return NS_OK;
}

NS_IMETHODIMP nsDrawingSurfaceNanoX::Unlock(void)
{
  if (!mLocked)
    return NS_ERROR_FAILURE;

  if (!(mLockFlags & NS_LOCK_SURFACE_READ_ONLY)) {
    GrArea(mDrawable, mGC, mLockX, mLockY, mLockWidth, mLockHeight,
           (void *)pixels, MWPF_PIXELVAL);
  }
  if (pixels) {
    free((void *)pixels);
    pixels = nsnull;
  }
  mLocked = PR_FALSE;
  return NS_OK;
}

NS_IMETHODIMP nsDrawingSurfaceNanoX::GetDimensions(PRUint32 *aWidth,
                                                   PRUint32 *aHeight)
{
  *aWidth = mWidth;
  *aHeight = mHeight;
  return NS_OK;
}

NS_IMETHODIMP nsDrawingSurfaceNanoX::IsOffscreen(PRBool *aOffScreen)
{
  *aOffScreen = mIsOffscreen;
  return NS_OK;
}

NS_IMETHODIMP nsDrawingSurfaceNanoX::IsPixelAddressable(PRBool *aAddressable)
{
  *aAddressable = PR_FALSE;
  return NS_OK;
}

NS_IMETHODIMP nsDrawingSurfaceNanoX::GetPixelFormat(nsPixelFormat *aFormat)
{
  *aFormat = mPixFormat;
  return NS_OK;
}

NS_IMETHODIMP nsDrawingSurfaceNanoX::SetRegion(GR_REGION_ID aRegion)
{
  if (mRegion)
    GrDestroyRegion(mRegion);

  mRegion = GrNewRegion();
  GrUnionRegion(mRegion, mRegion, aRegion);
  GrSetGCRegion(mGC, mRegion);
  return NS_OK;
}

PRUint8 nsDrawingSurfaceNanoX::ConvertMaskToCount(unsigned long val)
{
  PRUint8 retval = 0;
  PRUint8 cur_bit = 0;

  while (cur_bit < (sizeof(unsigned long) * 8)) {
    if ((val >> cur_bit) & 0x1) {
      retval++;
    }
    cur_bit++;
  }
  return retval;
}

PRUint8 nsDrawingSurfaceNanoX::GetShiftForMask(unsigned long val)
{
  PRUint8 cur_bit = 0;

  while (cur_bit < (sizeof(unsigned long) * 8)) {
    if ((val >> cur_bit) & 0x1) {
      return cur_bit;
    }
    cur_bit++;
  }
  return cur_bit;
}

