/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s):
 *   Peter Hartshorn <peter@igelaus.com.au>
 *   B.J. Rossiter <bj@igelaus.com.au>
 *   Matthew Wlazlo <mattw@igelaus.com.au>
 */

#include "nsRenderingContextNanoX.h"
#include "nsFontMetricsNanoX.h"
#include "nsICharRepresentable.h"
#include "prprf.h"
#include "prmem.h"

#include <nano-X.h>
#include <mwtypes.h>

#define NS_TO_GDK_RGB(ns) (ns & 0xff) << 16 | (ns & 0xff00) | ((ns >> 16) & 0xff)

static NS_DEFINE_IID(kIRenderingContextIID, NS_IRENDERING_CONTEXT_IID);

NS_IMPL_THREADSAFE_ISUPPORTS1(nsRenderingContextNanoX, nsIRenderingContext)

class GraphicsState
{
public:
	GraphicsState();
	~GraphicsState();

	nsTransform2D *mMatrix;
	nsRegionNanoX *mClipRegion;
	nscolor mColor;
	nsLineStyle mLineStyle;
	nsIFontMetrics *mFontMetrics;
};

GraphicsState::GraphicsState()
{
	mMatrix = nsnull;
	mClipRegion = nsnull;
	mColor = NS_RGB(255, 255, 255);
	mLineStyle = nsLineStyle_kSolid;
	mFontMetrics = nsnull;
}

GraphicsState::~GraphicsState()
{
	if (mClipRegion) {
		delete mClipRegion;
	}
	NS_IF_RELEASE(mFontMetrics);
}

nsRenderingContextNanoX::nsRenderingContextNanoX()
{
	NS_INIT_REFCNT();
	mOffscreenSurface = nsnull;
	mRenderingSurface = nsnull;
	mContext = nsnull;
	mFontMetrics = nsnull;
	mTranMatrix = nsnull;
	mP2T = 1.0f;
	mStateCache = new nsVoidArray();
	mCurrentFont = nsnull;
	mCurrentLineStyle = nsLineStyle_kSolid;
	mCurrentColor = NS_RGB(255, 255, 255);

	mDepth = 0;

	PushState();
}

nsRenderingContextNanoX::~nsRenderingContextNanoX()
{
	NS_IF_RELEASE(mContext);
	NS_IF_RELEASE(mFontMetrics);
	if (mStateCache) {
		PRInt32 cnt = mStateCache->Count();

		while (--cnt >= 0) {
			PRBool clipstate;
			PopState(clipstate);
		}
		delete mStateCache;
		mStateCache = nsnull;
	}
	if (mTranMatrix)
		delete mTranMatrix;

	if (mClipRegion) {
		delete mClipRegion;
	}
	NS_IF_RELEASE(mOffscreenSurface);
	NS_IF_RELEASE(mFontMetrics);
	NS_IF_RELEASE(mContext);

	if(mGC)
		GrDestroyGC(mGC);
}

NS_IMETHODIMP nsRenderingContextNanoX::Init(nsIDeviceContext *aContext, nsIWidget *aWindow)
{
	mContext = aContext;
	NS_IF_ADDREF(mContext);

	NS_ASSERTION(nsnull != mContext, "Device context is null.");


	mRenderingSurface = (nsDrawingSurfaceNanoX *)new nsDrawingSurfaceNanoX();

	if (mRenderingSurface) {
		if(!aWindow) return NS_ERROR_NULL_POINTER;
		GR_WINDOW_ID win = (GR_WINDOW_ID)aWindow->GetNativeData(NS_NATIVE_WINDOW);
		mRenderingSurface->Init(win);
		mOffscreenSurface = mRenderingSurface;
		NS_ADDREF(mRenderingSurface);
	}
	return CommonInit();
}

NS_IMETHODIMP nsRenderingContextNanoX::Init(nsIDeviceContext *aContext,
        nsDrawingSurface aSurface)
{
	mContext = aContext;
	NS_IF_ADDREF(mContext);

	NS_ASSERTION(nsnull != mContext, "Device context is null.");

	mRenderingSurface = (nsDrawingSurfaceNanoX *)aSurface;

	NS_IF_ADDREF(mRenderingSurface);

	return CommonInit();
}

nsresult nsRenderingContextNanoX::CommonInit()
{
	GR_WINDOW_INFO wInfo;
	GR_WINDOW_ID win = mRenderingSurface->GetDrawable();

	GrGetWindowInfo(win, &wInfo);

	if (mClipRegion) {
		delete mClipRegion;
	}

	mClipRegion = new nsRegionNanoX();
	mClipRegion->Init();
	mClipRegion->SetTo(0, 0, wInfo.width, wInfo.height);

	mContext->GetDevUnitsToAppUnits(mP2T);
	float app2dev;
	mContext->GetAppUnitsToDevUnits(app2dev);
	mTranMatrix->AddScale(app2dev, app2dev);

	mGC = GrNewGC();
	mGCChanged = true;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::Reset(void)
{
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetDeviceContext(nsIDeviceContext *&aContext)
{
	NS_IF_ADDREF(mContext);
	aContext = mContext;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::LockDrawingSurface(PRInt32 aX,
        PRInt32 aY, PRUint32 aWidth,
        PRUint32 aHeight, void **aBits,
        PRInt32 *aStride, PRInt32 *aWidthBytes,
        PRUint32 aFlags)
{
	PushState();
	return mRenderingSurface->Lock(aX, aY, aWidth, aHeight, aBits, aStride,
	                               aWidthBytes, aFlags);
}

NS_IMETHODIMP nsRenderingContextNanoX::UnlockDrawingSurface(void)
{
	PRBool clipstate;
	PopState(clipstate);
	mRenderingSurface->Unlock();
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::SelectOffScreenDrawingSurface(
    nsDrawingSurface aSurface)
{
	// XXX should copy gc's..
	if (nsnull == aSurface)
		mRenderingSurface = mOffscreenSurface;
	else
		mRenderingSurface = (nsDrawingSurfaceNanoX *)aSurface;

	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetDrawingSurface(
    nsDrawingSurface *aSurface)
{
	*aSurface = mRenderingSurface;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetHints(PRUint32 &aResult)
{
	PRUint32 result = 0;
	result |= NS_RENDERING_HINT_FAST_8BIT_TEXT;
	aResult = result;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::PushState(void)
{
	GraphicsState *state = new GraphicsState();

	if (!state)
		return NS_ERROR_OUT_OF_MEMORY;

	state->mMatrix = mTranMatrix;

	if (!mTranMatrix)
		mTranMatrix = new nsTransform2D();
	else
		mTranMatrix = new nsTransform2D(mTranMatrix);
	
//	mTranMatrix = mTranMatrix; // what the?!

	if (mClipRegion) {
		if (!state->mClipRegion) {
			state->mClipRegion = new nsRegionNanoX();
			state->mClipRegion->Init();
			state->mClipRegion->SetTo((const nsIRegion &)*(mClipRegion));
		} else 
			state->mClipRegion = mClipRegion;
	}
	NS_IF_ADDREF(mFontMetrics);
	state->mFontMetrics = mFontMetrics;
	state->mColor = mCurrentColor;
	state->mLineStyle = mCurrentLineStyle;

	mStateCache->AppendElement(state);
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::PopState(PRBool &aClipState)
{
	PRUint32 cnt = mStateCache->Count();
	GraphicsState *state;

	if (cnt > 0) {
		state = (GraphicsState *)mStateCache->ElementAt(cnt - 1);
		mStateCache->RemoveElementAt(cnt - 1);

		if(state->mMatrix) {
			if (mTranMatrix)
				delete mTranMatrix;
			mTranMatrix = state->mMatrix;
		}

		if (state->mClipRegion) {
			mClipRegion->SetTo((const nsIRegion &)*(state->mClipRegion));
		}

		if (state->mFontMetrics && (mFontMetrics != state->mFontMetrics))
			SetFont(state->mFontMetrics);

		if (mRenderingSurface && mClipRegion) {
			GR_REGION_ID region;
			mClipRegion->GetNativeRegion((void *&)region);
			//GrSetGCRegion(mGC, region);
			mRenderingSurface->SetRegion(region);
			GrDestroyRegion(region);
		}

		if (state->mColor != mCurrentColor)
			SetColor(state->mColor);
		if (state->mLineStyle != mCurrentLineStyle)
			SetLineStyle(state->mLineStyle);

		delete state;
	}

	if (mClipRegion)
		aClipState = mClipRegion->IsEmpty();
	else
		aClipState = PR_TRUE;

	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::IsVisibleRect(const nsRect &aRect,
        PRBool &aVisible)
{
	aVisible = PR_TRUE;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::SetClipRect(const nsRect &aRect,
        nsClipCombine aCombine,
        PRBool &aClipState)
{
	nsRect trect = aRect;
	GR_REGION_ID rgn;
	mTranMatrix->TransformCoord(&trect.x, &trect.y,
	                         &trect.width, &trect.height);
	switch (aCombine) {
		case nsClipCombine_kIntersect:
			mClipRegion->Intersect(trect.x, trect.y, trect.width, trect.height);
			break;
		case nsClipCombine_kUnion:
			mClipRegion->Union(trect.x, trect.y, trect.width, trect.height);
			break;
		case nsClipCombine_kSubtract:
			mClipRegion->Subtract(trect.x, trect.y, trect.width, trect.height);
			break;
		case nsClipCombine_kReplace:
			mClipRegion->SetTo(trect.x, trect.y, trect.width, trect.height);
			break;
	}
	aClipState = mClipRegion->IsEmpty();

	mClipRegion->GetNativeRegion((void*&)rgn);
	// GrSetGCRegion(mGC, rgn);
	mRenderingSurface->SetRegion(rgn);
	GrDestroyRegion(rgn);

	mGCChanged = true;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetClipRect(nsRect &aRect,
        PRBool &aClipState)
{
	PRInt32 x, y, w, h;
	if (!mClipRegion->IsEmpty()) {
		mClipRegion->GetBoundingBox(&x, &y, &w, &h);
		aRect.SetRect(x, y, w, h);
		aClipState = PR_TRUE;
	} else {
		aRect.SetRect(0, 0, 0, 0);
		aClipState = PR_FALSE;
	}
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::SetClipRegion(const nsIRegion &aRegion,
        nsClipCombine aCombine,
        PRBool &aClipState)
{
	GR_REGION_ID rgn;
	switch (aCombine) {
		case nsClipCombine_kIntersect:
			mClipRegion->Intersect(aRegion);
			break;
		case nsClipCombine_kUnion:
			mClipRegion->Union(aRegion);
			break;
		case nsClipCombine_kSubtract:
			mClipRegion->Subtract(aRegion);
			break;
		case nsClipCombine_kReplace:
			mClipRegion->SetTo(aRegion);
			break;
	}

	aClipState = mClipRegion->IsEmpty();
	mClipRegion->GetNativeRegion((void*&)rgn);
	//GrSetGCRegion(mGC, rgn);
	mRenderingSurface->SetRegion(rgn);
	GrDestroyRegion(rgn);
	mGCChanged = true;

	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::CopyClipRegion(nsIRegion &aRegion)
{
	aRegion.SetTo(*NS_STATIC_CAST(nsIRegion*, mClipRegion));
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetClipRegion(nsIRegion **aRegion)
{
	nsresult rv = NS_OK;

	NS_ASSERTION(!(nsnull == aRegion), "no region ptr");

	if (*aRegion) {
		nsIRegion *nRegion = (nsIRegion *)mClipRegion;
		(*aRegion)->SetTo(*nRegion);
	}
	return rv;
}

NS_IMETHODIMP nsRenderingContextNanoX::SetLineStyle(nsLineStyle aLineStyle)
{
	mCurrentLineStyle = aLineStyle;
	mGCChanged = true;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetLineStyle(nsLineStyle &aLineStyle)
{
	aLineStyle = mCurrentLineStyle;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::SetColor(nscolor aColor)
{
	if (nsnull == mContext)
		return NS_ERROR_FAILURE;
	mCurrentColor = aColor;
	mGCChanged = true;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetColor(nscolor &aColor) const
{
	aColor = mCurrentColor;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::SetFont(const nsFont &aFont)
{
	NS_IF_RELEASE(mFontMetrics);
	mContext->GetMetricsFor(aFont, mFontMetrics);
	return SetFont(mFontMetrics);
}

NS_IMETHODIMP nsRenderingContextNanoX::SetFont(nsIFontMetrics *aFontMetrics)
{
	NS_IF_RELEASE(mFontMetrics);
	mFontMetrics = aFontMetrics;
	NS_IF_ADDREF(mFontMetrics);

	if (mFontMetrics) {
		nsFontHandle fontHandle;
		mFontMetrics->GetFontHandle(fontHandle);
		mCurrentFont = (GR_FONT_ID)fontHandle;
		//GrSetGCFont(mGC, mCurrentFont);
		mGCChanged = true;
	}
	mGCChanged = true;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetFontMetrics(
    nsIFontMetrics *&aFontMetrics)
{
	NS_IF_ADDREF(mFontMetrics);
	aFontMetrics = mFontMetrics;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::Translate(nscoord aX, nscoord aY)
{
	mTranMatrix->AddTranslation((float)aX, (float)aY);
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::Scale(float aSx, float aSy)
{
	mTranMatrix->AddScale(aSx, aSy);
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetCurrentTransform(
    nsTransform2D *&aTransform)
{
	aTransform = mTranMatrix;
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::CreateDrawingSurface(nsRect *aBounds,
        PRUint32 aSurfFlags,
        nsDrawingSurface &aSurface)
{
	if (nsnull == mRenderingSurface) {
		aSurface = nsnull;
		return NS_ERROR_FAILURE;
	}

	nsDrawingSurfaceNanoX *surf = new nsDrawingSurfaceNanoX();

	if (surf) {
		NS_ADDREF(surf);
		UpdateGC();
		surf->Init(aBounds->width, aBounds->height, aSurfFlags);
	}

	aSurface = (nsDrawingSurface)surf;

	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DestroyDrawingSurface(
    nsDrawingSurface aDS)
{
	nsDrawingSurfaceNanoX *surf = (nsDrawingSurfaceNanoX *)aDS;

	NS_IF_RELEASE(surf);

	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawLine(nscoord aX0, nscoord aY0,
        nscoord aX1, nscoord aY1)
{
	nscoord diffX, diffY;

	if (nsnull == mTranMatrix || nsnull == mRenderingSurface)
		return NS_ERROR_FAILURE;

	mTranMatrix->TransformCoord(&aX0, &aY0);
	mTranMatrix->TransformCoord(&aX1, &aY1);

	diffX = aX1 - aX0;
	diffY = aY1 - aY0;

	if (0 != diffX) {
		diffX = (diffX > 0 ? 1 : -1);
	}
	if (0 != diffY) {
		diffY = (diffY > 0 ? 1 : -1);
	}

	UpdateGC();

	::GrLine(mRenderingSurface->GetDrawable(),
	         mGC, aX0, aY0, aX1 - diffX, aY1 - diffY);
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawStdLine(nscoord aX0, nscoord aY0,
        nscoord aX1, nscoord aY1)
{
	nscoord diffX, diffY;

	mTranMatrix->TransformCoord(&aX0, &aY0);
	mTranMatrix->TransformCoord(&aX1, &aY1);

	diffX = aX1 - aX0;
	diffY = aY1 - aY0;

	if (0 != diffX) {
		diffX = (diffX > 0 ? 1 : -1);
	}
	if (0 != diffY) {
		diffY = (diffY > 0 ? 1 : -1);
	}
	UpdateGC();
	::GrLine(mRenderingSurface->GetDrawable(),
	         mGC, aX0, aY0, aX1 - diffX, aY1 - diffY);
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawPolyline(const nsPoint aPoints[],
        PRInt32 aNumPoints)
{
	if (nsnull == mTranMatrix || nsnull == mRenderingSurface) {
		return NS_ERROR_FAILURE;
	}
	PRInt32 i ;
	GR_POINT *xpoints;
	GR_POINT *thispoint;

	xpoints = (GR_POINT *) PR_Malloc(sizeof(GR_POINT) * aNumPoints);

	for (i = 0; i < aNumPoints; i++) {
		thispoint = (xpoints + i);
		thispoint->x = aPoints[i].x;
		thispoint->y = aPoints[i].y;
		mTranMatrix->TransformCoord((PRInt32*)&thispoint->x, (PRInt32*)&thispoint->y);
	}

	UpdateGC();
	::GrPoly(mRenderingSurface->GetDrawable(), mGC,
	         aNumPoints, xpoints);

	PR_Free((void *)xpoints);

	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawRect(const nsRect &aRect)
{
	return DrawRect(aRect.x, aRect.y, aRect.width, aRect.height);
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawRect(nscoord aX, nscoord aY,
        nscoord aWidth, nscoord aHeight)
{
	if (nsnull == mTranMatrix || nsnull == mRenderingSurface)
		return NS_ERROR_FAILURE;

	nscoord x, y, w, h;

	x = aX;
	y = aY;
	w = aWidth;
	h = aHeight;

	mTranMatrix->TransformCoord(&x, &y, &w, &h);

	// Don't draw empty rectangles; also, w/h are adjusted down by one
	// so that the right number of pixels are drawn.
	if (w && h) {
		UpdateGC();
		GrRect(mRenderingSurface->GetDrawable(),
		       mGC,
		       x, y, w - 1, h - 1);
	}
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::FillRect(const nsRect &aRect)
{
	return FillRect(aRect.x, aRect.y, aRect.width, aRect.height);
}

NS_IMETHODIMP nsRenderingContextNanoX::FillRect(nscoord aX, nscoord aY,
        nscoord aWidth, nscoord aHeight)
{
	if (nsnull == mTranMatrix || nsnull == mRenderingSurface) {
		return NS_ERROR_FAILURE;
	}
	nscoord x, y, w, h;
	x = aX;
	y = aY;
	w = aWidth;
	h = aHeight;
	
	mTranMatrix->TransformCoord(&x, &y, &w, &h);

	//printf("****** nsRendContNanoX::FillRect\n");

	UpdateGC();
//	GR_COLOR c = MWRGB(NS_GET_R(mCurrentColor), NS_GET_G(mCurrentColor), NS_GET_B(mCurrentColor));
//	printf("FillRect: fg=0x%x, bg=0x%x\n", c, 0xffffff - c);
	::GrFillRect(mRenderingSurface->GetDrawable(),
	             mGC,
	             x, y, w, h);
	//printf("****** nsRendContNanoX::FillRect - DONE\n");
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::InvertRect(const nsRect& aRect)
{
	return InvertRect(aRect.x, aRect.y, aRect.width, aRect.height);
}

NS_IMETHODIMP nsRenderingContextNanoX::InvertRect(nscoord aX, nscoord aY,
        nscoord aWidth, nscoord aHeight)
{
	if (nsnull == mTranMatrix || nsnull == mRenderingSurface)
		return NS_ERROR_FAILURE;

	nscoord x, y, w, h;

	x = aX;
	y = aY;
	w = aWidth;
	h = aHeight;

	mTranMatrix->TransformCoord(&x, &y, &w, &h);

	UpdateGC();
	::GrSetGCMode(mGC, MWROP_SRCINVERT);
	::GrFillRect(mRenderingSurface->GetDrawable(),
	             mGC,
	             x, y, w, h);
	::GrSetGCMode(mGC, GR_MODE_SET);
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawPolygon(const nsPoint aPoints[],
        PRInt32 aNumPoints)
{
	if (nsnull == mTranMatrix || nsnull == mRenderingSurface) {
		return NS_ERROR_FAILURE;
	}
	PRInt32 i ;
	GR_POINT * xpoints;
	GR_POINT * thispoint;

	xpoints = (GR_POINT *) PR_Malloc(sizeof(GR_POINT) * aNumPoints);

	for (i = 0; i < aNumPoints; i++) {
		thispoint = (xpoints + i);
		thispoint->x = aPoints[i].x;
		thispoint->y = aPoints[i].y;
		mTranMatrix->TransformCoord((PRInt32*)&thispoint->x, (PRInt32*)&thispoint->y);
	}
	UpdateGC();
	::GrPoly(mRenderingSurface->GetDrawable(),
	         mGC,
	         aNumPoints, xpoints);
	PR_Free((void *)xpoints);
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::FillPolygon(const nsPoint aPoints[],
        PRInt32 aNumPoints)
{
	if (nsnull == mTranMatrix || nsnull == mRenderingSurface) {
		return NS_ERROR_FAILURE;
	}
	PRInt32 i ;
	GR_POINT * xpoints;
	GR_POINT * thispoint;
	nscoord x, y;

	xpoints = (GR_POINT *) PR_Malloc(sizeof(GR_POINT) * aNumPoints);

	for (i = 0; i < aNumPoints; i++) {
		thispoint = (xpoints + i);
		x = aPoints[i].x;
		y = aPoints[i].y;
		mTranMatrix->TransformCoord(&x, &y);
		thispoint->x = x;
		thispoint->y = y;
	}
	UpdateGC();
	::GrFillPoly(mRenderingSurface->GetDrawable(),
	             mGC,
	             aNumPoints, xpoints);
	PR_Free((void *)xpoints);
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawEllipse(const nsRect &aRect)
{
	return DrawEllipse(aRect.x, aRect.y, aRect.width, aRect.height);
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawEllipse(nscoord aX, nscoord aY,
        nscoord aWidth, nscoord aHeight)
{
	if (nsnull == mTranMatrix || nsnull == mRenderingSurface) {
		return NS_ERROR_FAILURE;
	}
	nscoord x, y, w, h;

	x = aX;
	y = aY;
	w = aWidth;
	h = aHeight;

	mTranMatrix->TransformCoord(&x, &y, &w, &h);
	UpdateGC();

	::GrEllipse(mRenderingSurface->GetDrawable(),
	            mGC,
	            x, y, w, h);
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::FillEllipse(const nsRect &aRect)
{
	return FillEllipse(aRect.x, aRect.y, aRect.width, aRect.height);
}

NS_IMETHODIMP nsRenderingContextNanoX::FillEllipse(nscoord aX, nscoord aY,
        nscoord aWidth, nscoord aHeight)
{
	if (nsnull == mTranMatrix || nsnull == mRenderingSurface) {
		return NS_ERROR_FAILURE;
	}
	nscoord x, y, w, h;

	x = aX;
	y = aY;
	w = aWidth;
	h = aHeight;

	mTranMatrix->TransformCoord(&x, &y, &w, &h);
	UpdateGC();

	::GrFillEllipse(mRenderingSurface->GetDrawable(),
	                mGC,
	                x, y, w, h);
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawArc(const nsRect &aRect,
        float aStartAngle,
        float aEndAngle)
{
	return DrawArc(aRect.x, aRect.y, aRect.width, aRect.height,
	               aStartAngle, aEndAngle);
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawArc(nscoord aX, nscoord aY,
        nscoord aWidth, nscoord aHeight,
        float aStartAngle, float aEndAngle)
{
	if (nsnull == mTranMatrix || nsnull == mRenderingSurface) {
		return NS_ERROR_FAILURE;
	}
	nscoord x, y, w, h;

	x = aX;
	y = aY;
	w = aWidth;
	h = aHeight;

	mTranMatrix->TransformCoord(&x, &y, &w, &h);
	UpdateGC();

	::GrArcAngle(mRenderingSurface->GetDrawable(),
	             mGC,
	             x, y, w, h, NSToIntRound(aStartAngle * 64.0f),
	             NSToIntRound(aEndAngle * 64.0f), GR_ARC);

	return NS_OK;
}

NS_IMETHODIMP
nsRenderingContextNanoX::FillArc(const nsRect& aRect,
                                 float aStartAngle, float aEndAngle)
{
	return FillArc(aRect.x, aRect.y, aRect.width, aRect.height,
	               aStartAngle, aEndAngle);
}

NS_IMETHODIMP nsRenderingContextNanoX::FillArc(nscoord aX, nscoord aY,
        nscoord aWidth, nscoord aHeight,
        float aStartAngle, float aEndAngle)
{
	if (nsnull == mTranMatrix || nsnull == mRenderingSurface) {
		return NS_ERROR_FAILURE;
	}
	nscoord x, y, w, h;

	x = aX;
	y = aY;
	w = aWidth;
	h = aHeight;

	mTranMatrix->TransformCoord(&x, &y, &w, &h);
	UpdateGC();

	::GrArcAngle(mRenderingSurface->GetDrawable(),
	             mGC,
	             x, y, w, h, NSToIntRound(aStartAngle * 64.0f),
	             NSToIntRound(aEndAngle * 64.0f), GR_PIE);

	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetWidth(char aC, nscoord& aWidth)
{
	// Check for the very common case of trying to get the width of a single
	// space.
	if ((aC == ' ') && (nsnull != mFontMetrics)) {
		nsFontMetricsNanoX* fontMetricsNanoX =
				(nsFontMetricsNanoX*)mFontMetrics;
		  return fontMetricsNanoX->GetSpaceWidth(aWidth);
	} 

	return GetWidth(&aC, 1, aWidth);
}

NS_IMETHODIMP nsRenderingContextNanoX::GetWidth(PRUnichar aC, nscoord& aWidth,
        PRInt32 *aFontID)
{
	return GetWidth(&aC, 1, aWidth, aFontID);
}

NS_IMETHODIMP nsRenderingContextNanoX::GetWidth(const nsString& aString,
        nscoord& aWidth, PRInt32 *aFontID)
{
	return GetWidth(aString.get(), aString.Length(), aWidth, aFontID);
}

NS_IMETHODIMP nsRenderingContextNanoX::GetWidth(const char* aString,
        nscoord& aWidth)
{
	return GetWidth(aString, strlen(aString), aWidth);
}

NS_IMETHODIMP nsRenderingContextNanoX::GetWidth(const char *aString,
        PRUint32 aLength, nscoord &aWidth)
{
	GR_FONT_INFO fInfo;

	if (!aLength)
		aWidth = 0;
	else {
		int rawWidth = 0;
		((nsFontMetricsNanoX*)mFontMetrics)->GetFontInfo(fInfo);
		for (unsigned int i = 0; i < aLength; i++) {
			rawWidth += fInfo.widths[aString[i]];
		}
		aWidth = NSToCoordRound(rawWidth * mP2T);
	}
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetWidth(const PRUnichar* aString, PRUint32 aLength,
        nscoord& aWidth, PRInt32 *aFontID)
{
	if (0 == aLength) {
		aWidth = 0;
	} else {
		nsFontMetricsNanoX* metrics = (nsFontMetricsNanoX*) mFontMetrics;
		GR_FONT_INFO fInfo;
		metrics->GetFontInfo(fInfo);
		int rawWidth = 0;
		for (unsigned int i = 0; i < aLength; i++)
			rawWidth += fInfo.widths[aString[i] % 256];

		aWidth = NSToCoordRound(rawWidth * mP2T);
	}
	if (nsnull != aFontID)
		*aFontID = 0;


	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawString(const char *aString,
        PRUint32 aLength,
        nscoord aX, nscoord aY,
        const nscoord *aSpacing)
{
	if (0 != aLength) {
		if (mTranMatrix == nsnull)
			return NS_ERROR_FAILURE;
		if (mRenderingSurface == nsnull)
			return NS_ERROR_FAILURE;
		if (aString == nsnull)
			return NS_ERROR_FAILURE;

		nscoord x = aX;
		nscoord y = aY;

		// Substract xFontStruct ascent since drawing specifies baseline
		if (mFontMetrics) {
			mFontMetrics->GetMaxAscent(y);
			y += aY;
		}

		UpdateGC();
		GR_GC_ID gc = GrCopyGC(mGC);
		GrSetGCUseBackground(gc, PR_FALSE);

		if (nsnull != aSpacing) {
			const char *end = aString + aLength;
			while (aString < end) {
				char ch = *aString++;
				nscoord xx = x;
				nscoord yy = y;
				mTranMatrix->TransformCoord(&xx, &yy);
				GrText(mRenderingSurface->GetDrawable(),
				       mGC,
				       xx, yy, &ch, 1, MWTF_ASCII);
				x += *aSpacing++;
			}
		} else {
			mTranMatrix->TransformCoord(&x, &y);
			GrText(mRenderingSurface->GetDrawable(),
			       gc,
			       x, y, (void *)aString, aLength, MWTF_ASCII);
		}
		GrDestroyGC(gc);
	}
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawString(const PRUnichar *aString,
        PRUint32 aLength,
        nscoord aX, nscoord aY,
        PRInt32 aFontID,
        const nscoord *aSpacing)
{
	if (aLength && mFontMetrics) {
		if (mTranMatrix == nsnull)
			return NS_ERROR_FAILURE;
		if (mRenderingSurface == nsnull)
			return NS_ERROR_FAILURE;
		if (aString == nsnull)
			return NS_ERROR_FAILURE;

		nscoord x = aX;
		nscoord y;

		// Substract xFontStruct ascent since drawing specifies baseline
		mFontMetrics->GetMaxAscent(y);
		y += aY;
		aY = y;

		mTranMatrix->TransformCoord(&x, &y);
		UpdateGC();

		nsFontMetricsNanoX* metrics = (nsFontMetricsNanoX*) mFontMetrics;
		GR_GC_ID gc = GrCopyGC(mGC);
		GrSetGCUseBackground(gc, PR_FALSE);

		GrSetGCFont(gc, metrics->mFontHandle);
		GrText(mRenderingSurface->GetDrawable(), gc,
		       x, y, (GR_CHAR *)aString, aLength, MWTF_UC16);
		GrDestroyGC(gc);

	} else
		printf("Havent got aLength of mFontMetrics!\n");

	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawString(const nsString& aString,
        nscoord aX, nscoord aY,
        PRInt32 aFontID,
        const nscoord* aSpacing)
{
	return DrawString(aString.get(), aString.Length(),
	                  aX, aY, aFontID, aSpacing);
}

#if 0
NS_IMETHODIMP nsRenderingContextNanoX::DrawImage(nsIImage *aImage, nscoord aX,
        nscoord aY)
{
	nscoord width, height;

	width = NSToCoordRound(mP2T * aImage->GetWidth());
	height = NSToCoordRound(mP2T * aImage->GetHeight());

	return DrawImage(aImage, aX, aY, width, height);
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawImage(nsIImage *aImage, nscoord aX,
        nscoord aY, nscoord aWidth,
        nscoord aHeight)
{
	nsRect tr;

	tr.x = aX;
	tr.y = aY;
	tr.width = aWidth;
	tr.height = aHeight;
	return DrawImage(aImage, tr);
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawImage(nsIImage *aImage,
        const nsRect &aRect)
{
	nsRect tr;
	tr = aRect;
	mTranMatrix->TransformCoord(&tr.x, &tr.y, &tr.width, &tr.height);
	nsresult res = aImage->Draw(*this, mRenderingSurface,
	                    tr.x, tr.y, tr.width, tr.height);
	return res;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawImage(nsIImage *aImage,
        const nsRect& aSRect,
        const nsRect& aDRect)
{
	nsRect sr, dr;

	sr = aSRect;
	mTranMatrix->TransformCoord(&sr.x, &sr.y,
	                         &sr.width, &sr.height);

	dr = aDRect;
	mTranMatrix->TransformCoord(&dr.x, &dr.y,
	                         &dr.width, &dr.height);

	return aImage->Draw(*this, mRenderingSurface,
	                    sr.x, sr.y,
	                    sr.width, sr.height,
	                    dr.x, dr.y,
	                    dr.width, dr.height);
}

#else

NS_IMETHODIMP nsRenderingContextNanoX::DrawImage(nsIImage *aImage, nscoord aX,
        nscoord aY)
{
	nscoord width, height;

	width = NSToCoordRound(mP2T * aImage->GetWidth());
	height = NSToCoordRound(mP2T * aImage->GetHeight());

	return DrawImage(aImage, aX, aY, width, height);
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawImage(nsIImage *aImage, nscoord aX,
        nscoord aY, nscoord aWidth,
        nscoord aHeight)
{

  nscoord x, y, w, h;                          
  
  x = aX;
  y = aY;
  w = aWidth;
  h = aHeight;
  
  mTranMatrix->TransformCoord(&x, &y, &w, &h);
	
  return aImage->Draw(*this, mRenderingSurface,
                      x, y, w, h);
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawImage(nsIImage *aImage,
        const nsRect &aRect)
{
  return DrawImage(aImage,
                   aRect.x,
                   aRect.y,
                   aRect.width,
                   aRect.height);
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawImage(nsIImage *aImage,
        const nsRect& aSRect,
        const nsRect& aDRect)
{
 nsRect        sr,dr;                         
  
  sr = aSRect;
  mTranMatrix->TransformCoord(&sr.x, &sr.y,
                            &sr.width, &sr.height);
  sr.x -= mTranMatrix->GetXTranslationCoord();
  sr.y -= mTranMatrix->GetYTranslationCoord();
  
  dr = aDRect;
  mTranMatrix->TransformCoord(&dr.x, &dr.y,
                           &dr.width, &dr.height);

  
  return aImage->Draw(*this, mRenderingSurface,
                      sr.x, sr.y,
                      sr.width, sr.height,
                      dr.x, dr.y,
                      dr.width, dr.height);
}

#endif
NS_IMETHODIMP nsRenderingContextNanoX::DrawTile(nsIImage *aImage,
        nscoord aSrcXOffset,
        nscoord aSrcYOffset,
        const nsRect &aTileRect)
{
	nsRect tileRect(aTileRect);
	nsRect srcRect(0, 0, aSrcXOffset, aSrcYOffset);
	mTranMatrix->TransformCoord(&srcRect.x, &srcRect.y, &srcRect.width,
	                         &srcRect.height);
	mTranMatrix->TransformCoord(&tileRect.x, &tileRect.y,
	                         &tileRect.width, &tileRect.height);

	if ((tileRect.width > 0) && (tileRect.height > 0))
		((nsImageNanoX*)aImage)->DrawTile(*this, mRenderingSurface, srcRect.width,
		                                  srcRect.height, tileRect);

	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::DrawTile(nsIImage *aImage,
        nscoord aX0,
        nscoord aY0,
        nscoord aX1,
        nscoord aY1,
        nscoord aWidth,
        nscoord aHeight)
{

	nsRect tileRect(aX0, aY0, aWidth, aHeight);
	nsRect srcRect(aX1, aY1, aWidth, aHeight);

	mTranMatrix->TransformCoord(&srcRect.x, &srcRect.y, &srcRect.width,
	                         &srcRect.height);
	mTranMatrix->TransformCoord(&tileRect.x, &tileRect.y,
	                         &tileRect.width, &tileRect.height);

	if ((tileRect.width > 0) && (tileRect.height > 0))
		((nsImageNanoX*)aImage)->DrawTile(*this,
		                                  mRenderingSurface,
		                                  srcRect.width,
		                                  srcRect.height, tileRect);
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::CopyOffScreenBits(
    nsDrawingSurface aSrcSurf, PRInt32 aSrcX,
    PRInt32 aSrcY, const nsRect &aDestBounds,
    PRUint32 aCopyFlags)
{
	PRInt32 srcX = aSrcX;
	PRInt32 srcY = aSrcY;
	nsRect drect = aDestBounds;
	nsDrawingSurfaceNanoX *destsurf;

	if (aSrcSurf == nsnull)
		return NS_ERROR_FAILURE;
	if (mTranMatrix == nsnull)
		return NS_ERROR_FAILURE;
	if (mRenderingSurface == nsnull)
		return NS_ERROR_FAILURE;

	if (aCopyFlags & NS_COPYBITS_TO_BACK_BUFFER) {
		NS_ASSERTION(!(nsnull == mRenderingSurface), "no back buffer");
		destsurf = mRenderingSurface;
	} else
		destsurf = mOffscreenSurface;

	if (aCopyFlags & NS_COPYBITS_XFORM_SOURCE_VALUES)
		mTranMatrix->TransformCoord(&srcX, &srcY);

	if (aCopyFlags & NS_COPYBITS_XFORM_DEST_VALUES)
		mTranMatrix->TransformCoord(&drect.x, &drect.y, &drect.width, &drect.height);

	UpdateGC();
	GrCopyArea(destsurf->GetDrawable(),
	           ((nsDrawingSurfaceNanoX *)aSrcSurf)->GetGC(),
	           drect.x, drect.y,
	           drect.width, drect.height,
	           ((nsDrawingSurfaceNanoX *)aSrcSurf)->GetDrawable(),
	           srcX, srcY, MWROP_SRCCOPY);

	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::RetrieveCurrentNativeGraphicData(
    PRUint32 *ngd)
{
	return NS_OK;
}

#ifdef MOZ_MATHML
NS_IMETHODIMP nsRenderingContextNanoX::GetBoundingMetrics(const char *aString,
        PRUint32 aLength,
        nsBoundingMetrics &aBoundingMetrics)
{
	return NS_OK;
}

NS_IMETHODIMP nsRenderingContextNanoX::GetBoundingMetrics(
    const PRUnichar *aString,
    PRUint32 aLength,
    nsBoundingMetrics &aBoundingMetrics,
    PRInt32 *aFontID)
{
	return NS_OK;
}

#endif

void nsRenderingContextNanoX::UpdateGC()
{
	if(!mGC) {
		mGC = GrNewGC();
		mGCChanged = true;
	}

	if(mGCChanged) {
		if(mClipRegion) {
			GR_REGION_ID rgn;
			mClipRegion->GetNativeRegion((void*&)rgn);
			GrSetGCRegion(mGC, rgn);
		}

		
		if(mCurrentColor) {
			GR_COLOR colour = MWRGB(NS_GET_R(mCurrentColor), NS_GET_G(mCurrentColor), NS_GET_B(mCurrentColor));
			GrSetGCForeground(mGC, colour);
			GrSetGCBackground(mGC, 0xffffff - colour);

			//printf("UpdateGC: Set fg: 0x%x, bg: 0x%x\n", color, 0xffffff - color);
		}
		
		if(mCurrentFont)
			GrSetGCFont(mGC, mCurrentFont);
	}
}
