/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s):
 *   Peter Hartshorn <peter@igelaus.com.au>
 *   B.J. Rossiter <bj@igelaus.com.au>
 *   Matthew Wlazlo <mattw@igelaus.com.au>
 */

#include <nsWidget.h>
#include <nsList.h>
#include <nsIServiceManager.h>
#include <nsIEventListener.h>
#include <nsIRenderingContext.h>
#include <nsIMenuListener.h>
#include <nsIMenuRollup.h>
#include <nsIRollupListener.h>
#include <nsIMouseListener.h>
#include <nsGfxCIID.h>
#include <nano-X.h>
#include <stdlib.h>
#include <stdio.h>
#include "bitmaps/cursors.h"

#define ARRSIZE(x) sizeof(x) / sizeof(x[0])

//#define DEBUG_mattw

#ifdef DEBUG_mattw
#	define ENTER {printf("Entering: %s\n",__PRETTY_FUNCTION__); }
#	define LEAVE {printf("Entering: %s\n",__PRETTY_FUNCTION__); }
#	define RETURN(c) { printf("Returning: 0x%x from %s\n",c,__PRETTY_FUNCTION__); return c; }
#else
#	define ENTER
#	define LEAVE
#	define RETURN(c) return c
#endif

static NS_DEFINE_CID(kRegionCID, NS_REGION_CID);

nsHashtable *nsWidget::gsWindowList = nsnull;
nsList *nsWidget::gsModalList = nsnull;

nsCOMPtr<nsIRollupListener> nsWidget::gRollupListener;
nsCOMPtr<nsIWidget> nsWidget::gRollupWidget;
PRBool             nsWidget::gRollupConsumeRollupEvent = PR_FALSE;

nsNanoXWindowCallback  nsWidget::gsWindowCreateCallback = nsnull;
nsNanoXWindowCallback  nsWidget::gsWindowDestroyCallback = nsnull;
nsNanoXEventDispatcher  nsWidget::gsEventDispatcher = nsnull;


class nsWindowKey : public nsHashKey {
protected:
  GR_WINDOW_ID mKey;

public:
  nsWindowKey(GR_WINDOW_ID key) {
    mKey = key;
  }
  ~nsWindowKey(void) {
  }
  PRUint32 HashCode(void) const {
    return (PRUint32)mKey;
  }
  PRBool Equals(const nsHashKey *aKey) const {
    return (mKey == ((nsWindowKey *)aKey)->mKey);
  }

  nsHashKey *Clone(void) const {
    return new nsWindowKey(mKey);
  }
};

GR_WINDOW_ID nsWidget::mFocusWindow = 0;

nsWidget *nsWidget::GetWidgetForWindow(GR_WINDOW_ID aWindow)
{
  if (gsWindowList == nsnull)
    return nsnull;

  nsWindowKey *window_key = new nsWindowKey(aWindow);
  nsWidget *retval = (nsWidget *)gsWindowList->Get(window_key);
  delete window_key;
  return retval;
}

void nsWidget::AddWindowCallback(GR_WINDOW_ID aWindow, nsWidget *aWidget)
{
  if (gsWindowList == nsnull)
    gsWindowList = new nsHashtable();

  nsWindowKey *window_key = new nsWindowKey(aWindow);
  gsWindowList->Put(window_key, aWidget);

  NS_ADDREF(aWidget);

  delete window_key;

  // make sure that if someone is listening that we inform
  // them of the new window
  if (gsWindowCreateCallback)
    (*gsWindowCreateCallback)(aWindow);

}

void nsWidget::DeleteWindowCallback(GR_WINDOW_ID aWindow)
{
  nsWindowKey *window_key = new nsWindowKey(aWindow);
  nsWidget *widget = (nsWidget *)gsWindowList->Get(window_key);
  NS_RELEASE(widget);
  gsWindowList->Remove(window_key);

  delete window_key;

  if(gsWindowDestroyCallback)
  	(*gsWindowDestroyCallback)(aWindow);
}

nsWidget::nsWidget() : nsBaseWidget()
{
  mBaseWindow = 0;
  mParentWindow = 0;
  mBackground = NS_RGB(192,192,192);
  mParentWidget = nsnull;
  mListenForResizes = PR_FALSE;
  mWindowType = eWindowType_child;
  mIsDestroying = PR_FALSE;
  mOnDestroyCalled = PR_FALSE;
  mIsShown = PR_FALSE;
  mIsToplevel = PR_FALSE;
  mName.AssignWithConversion("unnamed");

  mUpdateArea = do_CreateInstance(kRegionCID);
  if (mUpdateArea) {
    mUpdateArea->Init();
    mUpdateArea->SetTo(0,0,0,0);
  }
}

nsWidget::~nsWidget()
{
  if (mBaseWindow) {
    SetModal(PR_FALSE);
    Destroy();
  }
  mUpdateArea = nsnull;
}

NS_IMETHODIMP nsWidget::Create(nsIWidget *aParent,
                               const nsRect &aRect,
                               EVENT_CALLBACK aHandleEventFunction,
                               nsIDeviceContext *aContext,
                               nsIAppShell *aAppShell,
                               nsIToolkit *aToolkit,
                               nsWidgetInitData *aInitData)
{
  return (CommonCreate(aParent, aRect, aHandleEventFunction,
                      aContext, aAppShell, aToolkit, aInitData, nsnull));
}

NS_IMETHODIMP nsWidget::Create(nsNativeWidget aParent,
                               const nsRect &aRect,
                               EVENT_CALLBACK aHandleEventFunction,
                               nsIDeviceContext *aContext,
                               nsIAppShell *aAppShell,
                               nsIToolkit *aToolkit,
                               nsWidgetInitData *aInitData)
{
  return (CommonCreate(nsnull, aRect, aHandleEventFunction,
                      aContext, aAppShell, aToolkit, aInitData, aParent));
}

NS_IMETHODIMP nsWidget::CommonCreate(nsIWidget *aParent,
                                     const nsRect &aRect,
                                     EVENT_CALLBACK aHandleEventFunction,
                                     nsIDeviceContext *aContext,
                                     nsIAppShell *aAppShell,
                                     nsIToolkit *aToolkit,
                                     nsWidgetInitData *aInitData,
                                     nsNativeWidget aNativeParent)
{
  GR_WINDOW_ID parent=0;
  GR_WM_PROPERTIES prop;

  mBounds = aRect;

  if (mBounds.width <= 0) mBounds.width = 1;
  if (mBounds.height <= 0) mBounds.height = 1;

  mParentWidget = aParent;
  NS_IF_ADDREF(mParentWidget);

  BaseCreate(mParentWidget, mBounds, aHandleEventFunction, aContext,
             aAppShell, aToolkit, aInitData);

  if (aNativeParent) {
    parent = (GR_WINDOW_ID)aNativeParent;
    mListenForResizes = PR_TRUE;
  } else if (aParent) {
    parent = (GR_WINDOW_ID)aParent->GetNativeData(NS_NATIVE_WINDOW);
  } else {
    parent = GR_ROOT_WINDOW_ID;
  }

  if (nsnull != aInitData) {
    mWindowType = aInitData->mWindowType;
  }

  mParentWindow = parent;
  prop.props = 0;

  switch (mWindowType) {
  case eWindowType_popup:
    mIsToplevel = PR_TRUE;
    parent = GR_ROOT_WINDOW_ID;
    prop.props |= GR_WM_PROPS_NODECORATE | GR_WM_PROPS_NOAUTOMOVE |
                  GR_WM_PROPS_NOFOCUS;
    mBaseWindow = GrNewWindow(parent, mBounds.x, mBounds.y,
                              mBounds.width, mBounds.height, 0,
                              mBackground, 0x00ffff);
    AddWindowCallback(mBaseWindow, this);
    break;
  case eWindowType_dialog:
  case eWindowType_toplevel:
    mIsToplevel = PR_TRUE;
    parent = GR_ROOT_WINDOW_ID;
    mBaseWindow = GrNewWindow(parent, mBounds.x, mBounds.y,
                              mBounds.width, mBounds.height, 0,
                              mBackground, 0xffffff);
    AddWindowCallback(mBaseWindow, this);
    break;
  case eWindowType_child:
    mIsToplevel = PR_FALSE;
    mBaseWindow = GrNewWindow(parent, mBounds.x, mBounds.y,
                              mBounds.width, mBounds.height, 0,
                              mBackground, 0x00ff00);
    AddWindowCallback(mBaseWindow, this);
    break;
  default:
    break;
  }
  prop.flags = GR_WM_FLAGS_PROPS;
  prop.props |= GR_WM_PROPS_NOBACKGROUND;
  GrSetWMProperties(mBaseWindow, &prop);
  GrSelectEvents(mBaseWindow, GR_EVENT_MASK_ALL);
  mRequestedSize.x = mBounds.x;
  mRequestedSize.y = mBounds.y;
  mRequestedSize.height = mBounds.height;
  mRequestedSize.width = mBounds.width;

  return NS_OK;
}

void nsWidget::DestroyNative(void)
{
  if (mBaseWindow) {
    GrDestroyWindow(mBaseWindow);
    DeleteWindowCallback(mBaseWindow);
  }
}

NS_IMETHODIMP nsWidget::Destroy()
{
  if (mIsDestroying)
    return NS_OK;

  mIsDestroying = PR_TRUE;

  nsBaseWidget::Destroy();
  NS_IF_RELEASE(mParentWidget);

  if (mBaseWindow) {
    DestroyNative();
    if (PR_FALSE == mOnDestroyCalled)
      OnDestroy();
    mBaseWindow = nsnull;
    mEventCallback = nsnull;
  }

  return NS_OK;
}

NS_IMETHODIMP nsWidget::ConstrainPosition(PRInt32 *aX, PRInt32 *aY)
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::Move(PRInt32 aX, PRInt32 aY)
{
  mBounds.x = aX;
  mBounds.y = aY;

  if (mWindowType == eWindowType_popup) {
    nsRect aRect, transRect;

    aRect.x = aX;
    aRect.y = aY;

    if (mParentWidget) {
      mParentWidget->WidgetToScreen(aRect, transRect);
    } else {
      WidgetToScreen(aRect, transRect);
    }

    aX = transRect.x;
    aY = transRect.y;
  }

  mRequestedSize.x = aX;
  mRequestedSize.y = aY;

  if (mParentWidget) {
    ((nsWidget *)mParentWidget)->WidgetMove(this);
  } else {
    GrMoveWindow(mBaseWindow, aX, aY);
  }

  return NS_OK;
}

NS_IMETHODIMP nsWidget::Resize(PRInt32 aWidth, PRInt32 aHeight, PRBool aRepaint)
{
  if (aWidth <= 0) aWidth = 1;
  if (aHeight <= 0) aHeight = 1;

  mRequestedSize.width = mBounds.width = aWidth;
  mRequestedSize.height = mBounds.height = aHeight;

  if (mParentWidget) {
    ((nsWidget *)mParentWidget)->WidgetResize(this);
  } else {
    GrResizeWindow(mBaseWindow, aWidth, aHeight);
  }
  return NS_OK;
}

NS_IMETHODIMP nsWidget::Resize(PRInt32 aX, PRInt32 aY,
                               PRInt32 aWidth, PRInt32 aHeight,
                               PRBool aRepaint)
{

  if (aWidth <= 0) aWidth = 1;
  if (aHeight <= 0) aHeight = 1;

  if (aX < 0) aX = 0;
  if (aY < 0) aY = 0;

  mRequestedSize.x = aX;
  mRequestedSize.y = aY;
  mRequestedSize.width = mBounds.width = aWidth;
  mRequestedSize.height = mBounds.height = aHeight;

  if (mParentWidget) {
    ((nsWidget *)mParentWidget)->WidgetMoveResize(this);
  } else {
    GrMoveWindow(mBaseWindow, aX, aY);
    GrResizeWindow(mBaseWindow, aWidth, aHeight);
  }
  return NS_OK;
}

NS_IMETHODIMP nsWidget::Enable(PRBool bState)
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::SetFocus(PRBool aRaise)
{
  if(mBaseWindow)
    mFocusWindow = mBaseWindow;
  return NS_OK;
}

NS_IMETHODIMP nsWidget::SetName(const char *aName)
{
  mName.AssignWithConversion(aName);
  return NS_OK;
}

GR_WINDOW_ID nsWidget::GetFocusWindow(void)
{
  return mFocusWindow;
}

NS_IMETHODIMP nsWidget::Invalidate(PRBool aIsSynchronous)
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::Invalidate(const nsRect &aRect, PRBool aIsSyncronous)
{
  return NS_OK;
}

nsIFontMetrics* nsWidget::GetFont(void)
{
  return nsnull;
}

NS_IMETHODIMP nsWidget::GetPreferredSize(PRInt32 &aWidth, PRInt32 &aHeight)
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::SetMenuBar(nsIMenuBar * aMenuBar)
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::ShowMenuBar(PRBool aShow)
{
  return NS_OK;
}

void *nsWidget::GetNativeData(PRUint32 aDataType)
{
  switch (aDataType) {
  case NS_NATIVE_WIDGET:
  case NS_NATIVE_WINDOW:
  case NS_NATIVE_PLUGIN_PORT:
    return (void *)mBaseWindow;
    break;
  default:
    return 0;
    break;
  }
}

NS_IMETHODIMP nsWidget::SetTooltips(PRUint32 aNumberOfTips,
                                    nsRect* aTooltipAreas[])
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::UpdateTooltips(nsRect* aNewTips[])
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::RemoveTooltips()
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::SetFont(const nsFont &aFont)
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::BeginResizingChildren(void)
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::EndResizingChildren(void)
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::SetColorMap(nsColorMap *aColorMap)
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::SetModal(PRBool aModal)
{
  if (aModal) {
    if (!gsModalList) gsModalList = new nsList();
#ifdef DEBUG
    fprintf(stderr, "SetModal 1 this = 0x%x\n", this);
#endif
    gsModalList->add((void *)this);
  } else {
    if (!gsModalList) return NS_OK;

#ifdef DEBUG
fprintf(stderr, "SetModal 0 this = 0x%x\n", this);
#endif
    gsModalList->remove((void *)this);
  }
  return NS_OK;
}

NS_IMETHODIMP nsWidget::Show(PRBool bState)
{
  if (bState) {
    Map();
    mIsShown = PR_TRUE;
  } else {
    Unmap();
    mIsShown = PR_FALSE;
  }
  return NS_OK;
}

NS_IMETHODIMP nsWidget::IsVisible(PRBool &aState)
{
  aState = mIsShown;
  return NS_OK;
}

NS_IMETHODIMP nsWidget::SetPreferredSize(PRInt32 aWidth, PRInt32 aHeight)
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::Update()
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::SetBackgroundColor(const nscolor &aColor)
{
  GR_WM_PROPERTIES props;
  props.flags = GR_WM_FLAGS_BACKGROUND;
  props.background = MWRGB(NS_GET_R(aColor),
                           NS_GET_G(aColor),
                           NS_GET_B(aColor));
  nsBaseWidget::SetBackgroundColor(aColor);
  GrSetWMProperties(mBaseWindow, &props);
  return NS_OK;
}

NS_IMETHODIMP nsWidget::Scroll(PRInt32 aDx, PRInt32 aDy, nsRect *aClipRect)
{
  return NS_OK;
}

NS_IMETHODIMP nsWidget::WidgetToScreen(const nsRect &aOldRect,
                                       nsRect &aNewRect)
{
  GR_WINDOW_INFO wInfo;
  GrGetWindowInfo(mBaseWindow, &wInfo);
  aNewRect.x = wInfo.x + aOldRect.x;
  aNewRect.y = wInfo.y + aOldRect.y;

  return NS_OK;
}

NS_IMETHODIMP nsWidget::ScreenToWidget(const nsRect &aOldRect,
                                       nsRect &aNewRect)
{
  GR_WINDOW_INFO wInfo;
  GrGetWindowInfo(mBaseWindow, &wInfo);
  aNewRect.x = aOldRect.x - wInfo.x;
  aNewRect.y = aOldRect.y - wInfo.y;

  return NS_OK;
}

NS_IMETHODIMP nsWidget::SetCursor(nsCursor aCursor)
{
  MWIMAGEBITS bitmapfg[16];
  MWIMAGEBITS bitmapbg[16];
  int hotx, hoty;
	
	if(!mBaseWindow)
		return NS_ERROR_FAILURE;
	
  // dont re-load a cursor that is alread active
  if(aCursor != mCursor) {
  		mCursor = aCursor;

	  switch (aCursor) {
		case eCursor_standard: {
			LoadCursor(left_arrow_bits, ARRSIZE(left_arrow_bits), bitmapfg);
			LoadCursor(left_arrow_bg_bits, ARRSIZE(left_arrow_bg_bits), bitmapbg);
		  hotx = left_arrow_x_hot;
		  hoty = left_arrow_y_hot;
		}
		break;
		case eCursor_wait:
		case eCursor_spinning:
			LoadCursor(wait_bits, ARRSIZE(wait_bits), bitmapfg);
			LoadCursor(wait_bg_bits, ARRSIZE(wait_bg_bits), bitmapbg);
		  hotx = wait_x_hot;
		  hoty = wait_y_hot;

		break;
		case eCursor_select:
			LoadCursor(select_bits, ARRSIZE(select_bits), bitmapfg);
			LoadCursor(select_bg_bits, ARRSIZE(select_bg_bits), bitmapbg);
			hotx = select_x_hot;
			hoty = select_y_hot;
		break;
		case eCursor_hyperlink:
			LoadCursor(hyperlink_bits, ARRSIZE(hyperlink_bits), bitmapfg);
			LoadCursor(hyperlink_bg_bits, ARRSIZE(hyperlink_bg_bits), bitmapbg);
			hotx = hyperlink_x_hot;
			hoty = hyperlink_y_hot;
		break;
		case eCursor_sizeWE:
			LoadCursor(sizeWE_bits, ARRSIZE(sizeWE_bits), bitmapfg);
			LoadCursor(sizeWE_bg_bits, ARRSIZE(sizeWE_bg_bits), bitmapbg);
			hotx = sizeWE_x_hot;
			hoty = sizeWE_y_hot;
		break;
		
		case eCursor_sizeNS:
			LoadCursor(sizeNS_bits, ARRSIZE(sizeNS_bits), bitmapfg);
			LoadCursor(sizeNS_bg_bits, ARRSIZE(sizeNS_bg_bits), bitmapbg);
			hotx = sizeNS_x_hot;
			hoty = sizeNS_y_hot;
		break;

		case eCursor_sizeNE:
			LoadCursor(sizeNE_bits, ARRSIZE(sizeNE_bits), bitmapfg);
			LoadCursor(sizeNE_bg_bits, ARRSIZE(sizeNE_bg_bits), bitmapbg);
			hotx = sizeNE_x_hot;
			hoty = sizeNE_y_hot;
		break;

		case eCursor_sizeNW:
			LoadCursor(sizeNW_bits, ARRSIZE(sizeNW_bits), bitmapfg);
			LoadCursor(sizeNW_bg_bits, ARRSIZE(sizeNW_bg_bits), bitmapbg);
			hotx = sizeNW_x_hot;
			hoty = sizeNW_y_hot;
		break;

		case eCursor_sizeSE:
			LoadCursor(sizeSE_bits, ARRSIZE(sizeSE_bits), bitmapfg);
			LoadCursor(sizeSE_bg_bits, ARRSIZE(sizeSE_bg_bits), bitmapbg);
			hotx = sizeSE_x_hot;
			hoty = sizeSE_y_hot;
		break;
		
		case eCursor_sizeSW:
			LoadCursor(sizeSW_bits, ARRSIZE(sizeSW_bits), bitmapfg);
			LoadCursor(sizeSW_bg_bits, ARRSIZE(sizeSW_bg_bits), bitmapbg);
			hotx = sizeSW_x_hot;
			hoty = sizeSW_y_hot;
		break;

		case eCursor_crosshair:
			LoadCursor(crosshair_bits, ARRSIZE(crosshair_bits), bitmapfg);
			LoadCursor(crosshair_bg_bits, ARRSIZE(crosshair_bg_bits), bitmapbg);
			hotx = crosshair_x_hot;
			hoty = crosshair_y_hot;
		break;

		case eCursor_move:
			LoadCursor(move_bits, ARRSIZE(crosshair_bits), bitmapfg);
			LoadCursor(move_bg_bits, ARRSIZE(crosshair_bg_bits), bitmapbg);
			hotx = move_x_hot;
			hoty = move_y_hot;
		break;

      case eCursor_grab:
			LoadCursor(grab_bits, ARRSIZE(grab_bits), bitmapfg);
			LoadCursor(grab_bg_bits, ARRSIZE(grab_bg_bits), bitmapbg);
			hotx = grab_x_hot;
			hoty = grab_y_hot;
		break;

      case eCursor_grabbing:
			LoadCursor(grabbing_bits, ARRSIZE(grabbing_bits), bitmapfg);
			LoadCursor(grabbing_bg_bits, ARRSIZE(grabbing_bg_bits), bitmapbg);
			hotx = grabbing_x_hot;
			hoty = grabbing_y_hot;
		break;

      case eCursor_copy:    
			LoadCursor(copy_bits, ARRSIZE(copy_bits), bitmapfg);
			LoadCursor(copy_bg_bits, ARRSIZE(copy_bg_bits), bitmapbg);
			hotx = copy_x_hot;
			hoty = copy_y_hot;
		break;

	  case eCursor_arrow_south:
	  case eCursor_arrow_south_plus:
			LoadCursor(arrowSouth_bits, ARRSIZE(arrowSouth_bits), bitmapfg);
			LoadCursor(arrowSouth_bg_bits, ARRSIZE(arrowSouth_bg_bits), bitmapbg);
			hotx = arrowSouth_x_hot;
			hoty = arrowSouth_y_hot;
		break;
	  case eCursor_arrow_north:
	  case eCursor_arrow_north_plus:
			LoadCursor(arrowNorth_bits, ARRSIZE(arrowNorth_bits), bitmapfg);
			LoadCursor(arrowNorth_bg_bits, ARRSIZE(arrowNorth_bg_bits), bitmapbg);
			hotx = arrowNorth_x_hot;
			hoty = arrowNorth_y_hot;
		break;
	  case eCursor_arrow_west:
	  case eCursor_arrow_west_plus:
			LoadCursor(arrowWest_bits, ARRSIZE(arrowWest_bits), bitmapfg);
			LoadCursor(arrowWest_bg_bits, ARRSIZE(arrowWest_bg_bits), bitmapbg);
			hotx = arrowWest_x_hot;
			hoty = arrowWest_y_hot;
		break;
	  case eCursor_arrow_east:
	  case eCursor_arrow_east_plus:
			LoadCursor(arrowEast_bits, ARRSIZE(arrowEast_bits), bitmapfg);
			LoadCursor(arrowEast_bg_bits, ARRSIZE(arrowEast_bg_bits), bitmapbg);
			hotx = arrowEast_x_hot;
			hoty = arrowEast_y_hot;
		break;

      // CSS3 - see http://www.w3.org/TR/css3-userint#cursor
      case eCursor_alias:   
      case eCursor_context_menu:
      case eCursor_cell:
      case eCursor_count_up:
      case eCursor_count_down:
      case eCursor_count_up_down:
		default:	// also eCursor_help
			LoadCursor(default_bits, ARRSIZE(default_bits), bitmapfg);
			LoadCursor(default_bg_bits, ARRSIZE(default_bg_bits), bitmapbg);
			hotx = default_x_hot;
			hoty = default_y_hot;
		break;
	  }

  	GrSetCursor(mBaseWindow, 16, 16, hotx, hoty,
              0xffffff, 0x0, bitmapfg, bitmapbg);
  	GrSetCursor(mParentWindow, 16, 16, hotx, hoty,
              0xffffff, 0x0, bitmapfg, bitmapbg);
  }
  return NS_OK;
}

nsIWidget *nsWidget::GetParent(void)
{
  NS_IF_ADDREF(mParentWidget);
  return mParentWidget;
}

void nsWidget::WidgetMove(nsWidget *aWidget)
{
  GrMoveWindow(aWidget->mBaseWindow,
              aWidget->mRequestedSize.x, aWidget->mRequestedSize.y);
}

void nsWidget::WidgetResize(nsWidget *aWidget)
{
  GrResizeWindow(aWidget->mBaseWindow,
                aWidget->mRequestedSize.width,
                aWidget->mRequestedSize.height);
}

void nsWidget::WidgetMoveResize(nsWidget *aWidget)
{
  GrMoveWindow(aWidget->mBaseWindow,
              aWidget->mRequestedSize.x, aWidget->mRequestedSize.y);
  GrResizeWindow(aWidget->mBaseWindow,
                aWidget->mRequestedSize.width,
                aWidget->mRequestedSize.height);
}

void nsWidget::WidgetShow(nsWidget *aWidget)
{
  aWidget->Map();
}

void nsWidget::Map(void)
{
  GrMapWindow(mBaseWindow);
}

void nsWidget::Unmap(void)
{
  GrUnmapWindow(mBaseWindow);
}

NS_IMETHODIMP nsWidget::SetTitle(const nsString &title)
{
#ifdef DEBUG_mattw
  printf("*************nsWidget::SetTitle %s\n", title.GetBuffer());
#endif
  GR_WM_PROPERTIES props;

  props.title = (GR_CHAR*)title.ToNewCString();
  props.flags = GR_WM_FLAGS_TITLE;

  GrSetWMProperties(mBaseWindow, &props);
  return NS_OK;
}

NS_IMETHODIMP nsWidget::DispatchEvent(nsGUIEvent *aEvent,
                                      nsEventStatus &aStatus)
{
  NS_ADDREF(aEvent->widget);

  aStatus = nsEventStatus_eIgnore;
  if (mEventCallback) {
    aStatus = (*mEventCallback)(aEvent);
   }

  if ((aStatus != nsEventStatus_eIgnore) && (mEventListener)) {
    aStatus = mEventListener->ProcessEvent(*aEvent);
   }
  NS_RELEASE(aEvent->widget);
  return(NS_OK);
}

PRBool nsWidget::DispatchWindowEvent(nsGUIEvent &aEvent)
{
  nsEventStatus status;
  DispatchEvent(&aEvent, status);
  return ConvertStatus(status);
}

PRBool nsWidget::IsMouseInWindow(nsIWidget* inWidget, PRInt32 inMouseX, PRInt32 inMouseY){
  GR_WINDOW_INFO focusWindowInfo;
  GR_WINDOW_INFO thisWidgetInfo;  
  
  GR_WINDOW_ID currentPopupWindow = (GR_WINDOW_ID)inWidget->GetNativeData(NS_NATIVE_WINDOW);      
  GrGetWindowInfo(currentPopupWindow, &focusWindowInfo);
  
  GrGetWindowInfo(mBaseWindow, &thisWidgetInfo);
  
  // Test using coordinates (relative to root window) if click was inside passed popup window
  if ((thisWidgetInfo.x+inMouseX) > focusWindowInfo.x &&
      (thisWidgetInfo.x+inMouseX) < (focusWindowInfo.x + focusWindowInfo.width) &&
      (thisWidgetInfo.y+inMouseY) > focusWindowInfo.y &&
      (thisWidgetInfo.y+inMouseY) < (focusWindowInfo.y + focusWindowInfo.height)){
    return PR_TRUE;
  }
  return PR_FALSE;
}

PRBool nsWidget::HandlePopup ( PRInt32 inMouseX, PRInt32 inMouseY )
{
  PRBool retVal = PR_FALSE;
  PRBool rollup = PR_FALSE;

  // The gRollupListener and gRollupWidget are both set to nsnull when 
  // a popup is no longer visible

//  nsCOMPtr<nsIWidget> rollupWidget = do_QueryReferent(gRollupWidget);

  if (gRollupWidget && gRollupListener) {
//    GR_WINDOW_ID wid = (GR_WINDOW_ID)rollupWidget->GetNativeData(NS_NATIVE_WINDOW);

    if (!IsMouseInWindow(gRollupWidget, inMouseX, inMouseY)) {
      rollup = PR_TRUE;
      nsCOMPtr<nsIMenuRollup> menuRollup ( do_QueryInterface(gRollupListener) );

      if ( menuRollup ) {
        nsCOMPtr<nsISupportsArray> widgetChain;
        menuRollup->GetSubmenuWidgetChain ( getter_AddRefs(widgetChain) );
	
        if ( widgetChain ) {
          PRUint32 count = 0;
          widgetChain->Count ( &count );

          for ( PRUint32 i = 0; i < count; ++i ) {
            nsCOMPtr<nsISupports> genericWidget;
            widgetChain->GetElementAt ( i, getter_AddRefs(genericWidget) );
            nsCOMPtr<nsIWidget> widget ( do_QueryInterface(genericWidget) );

            if ( widget ) {
              if ( IsMouseInWindow(widget, inMouseX, inMouseY) ) {
                rollup = PR_FALSE;
                break;
              }
            }
          } // foreach parent menu widget
        }
      }       
    }                                       
  }
        
  if (rollup){
    gRollupListener->Rollup();
    retVal = PR_TRUE;
  }
  return retVal;
}


PRBool nsWidget::DispatchMouseEvent(nsMouseEvent &aEvent)
{
  PRBool result = PR_FALSE;

  if(nsnull == mEventCallback && nsnull == mMouseListener) 
  	return result;

  // If there was a mouse down event, check if any popups need to be informed
  switch (aEvent.message) {
    case NS_MOUSE_LEFT_BUTTON_DOWN:
    case NS_MOUSE_MIDDLE_BUTTON_DOWN:
    case NS_MOUSE_RIGHT_BUTTON_DOWN:                    
    if (HandlePopup(aEvent.point.x, aEvent.point.y)){
      return PR_TRUE;
    }
    break;
  }
  
  if (nsnull != mEventCallback) {
    return DispatchWindowEvent(aEvent);
  }
  if (nsnull != mMouseListener) {
    switch (aEvent.message) {
    case NS_MOUSE_MOVE:
      break;
    case NS_MOUSE_LEFT_BUTTON_DOWN:
    case NS_MOUSE_MIDDLE_BUTTON_DOWN:
    case NS_MOUSE_RIGHT_BUTTON_DOWN:
      result = ConvertStatus(mMouseListener->MousePressed(aEvent));
      break;
    case NS_MOUSE_LEFT_BUTTON_UP:
    case NS_MOUSE_MIDDLE_BUTTON_UP:
    case NS_MOUSE_RIGHT_BUTTON_UP:
      result = ConvertStatus(mMouseListener->MouseReleased(aEvent));
      result = ConvertStatus(mMouseListener->MouseClicked(aEvent));
      break;
    }
  }
  return result;
}

PRBool nsWidget::DispatchKeyEvent(nsKeyEvent &aKeyEvent)
{
  if (mEventCallback)
    return DispatchWindowEvent(aKeyEvent);
  return PR_FALSE;
}

PRBool nsWidget::DispatchDestroyEvent(void)
{
  PRBool result = PR_FALSE;
  if (nsnull != mEventCallback) {
    nsGUIEvent event;
    event.eventStructType = NS_GUI_EVENT;
    event.message = NS_DESTROY;
    event.widget = this;
    event.time = PR_Now();
    event.point.x = 0;
    event.point.y = 0;
    AddRef();
    result = DispatchWindowEvent(event);
    Release();
  }
  return result;
}

PRBool nsWidget::ConvertStatus(nsEventStatus aStatus)
{
  switch (aStatus) {
    case nsEventStatus_eIgnore:
      return(PR_FALSE);
    case nsEventStatus_eConsumeNoDefault:
      return(PR_TRUE);
    case nsEventStatus_eConsumeDoDefault:
      return(PR_FALSE);
    default:
      break;
  }
  return(PR_TRUE);
}

NS_IMETHODIMP nsWidget::SetBounds(const nsRect &aRect)
{
  mRequestedSize = aRect;
  return nsBaseWidget::SetBounds(aRect);
}

NS_IMETHODIMP nsWidget::GetRequestedBounds(nsRect &aRect)
{
  aRect = mRequestedSize;
  return NS_OK;
}

PRBool nsWidget::OnPaint(nsPaintEvent &event)
{
  nsresult result = PR_FALSE;
  if (mEventCallback) {
    event.renderingContext = GetRenderingContext();
    result = DispatchWindowEvent(event);
    NS_RELEASE(event.renderingContext);
  }
  return(result);
}

PRBool nsWidget::OnResize(nsSizeEvent &event)
{
  mBounds.width = event.mWinWidth;
  mBounds.height = event.mWinHeight;

  nsresult result = PR_FALSE;
  if (mEventCallback) {
      result = DispatchWindowEvent(event);
  }
  return result;
}

void nsWidget::OnDestroy()
{
  mOnDestroyCalled = PR_TRUE;
  nsBaseWidget::OnDestroy();
  DispatchDestroyEvent();
}

PRBool nsWidget::OnDeleteWindow()
{
  nsBaseWidget::OnDestroy();
  return DispatchDestroyEvent();
}

NS_IMETHODIMP nsWidget::CaptureRollupEvents(nsIRollupListener *aListener,
                                            PRBool aDoCapture,
                                            PRBool aConsumeRollupEvent)
{
  return NS_OK;
}

nsWidget *nsWidget::GetTopLevelWidget()
{
  GR_WINDOW_INFO wInfo;
  nsWidget *tlWidget = nsnull;

  GrGetWindowInfo(mBaseWindow, &wInfo);
  while ((wInfo.parent != GR_ROOT_WINDOW_ID) && (nsWidget::GetWidgetForWindow(wInfo.parent) != nsnull)) {
    GrGetWindowInfo(wInfo.parent, &wInfo);
  }

  tlWidget = nsWidget::GetWidgetForWindow(wInfo.wid);
  return tlWidget;
}

nsWidget *nsWidget::GetModalWindow()
{
  if (gsModalList) {
    nsListItem *aItem = gsModalList->getTail();
    if (aItem) return (nsWidget*)aItem->getData();
  }
  return nsnull;
}

nsWindowType nsWidget::GetWindowType()
{
  return mWindowType;
}

void nsWidget::LoadCursor(const unsigned char *bits, const int size, MWIMAGEBITS *out)
{
	int idx=0;
	for(int i=0; i < size; i+=2) {
		out[idx] = (bits[i] << 8) | bits[i+1];
		idx++;
	}
}
