/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s):
 *   Peter Hartshorn <peter@igelaus.com.au>
 *   B.J. Rossiter <bj@igelaus.com.au>
 */

#include <nsWindow.h>
#include <nsList.h>
#include <nsFileSpec.h>

#include <stdlib.h>
#include <stdio.h>

#define ABS(i) ( i<0 ? 0-i : i )

static nsList *update_queue = nsnull;

void
nsWindow::UpdateIdle()
{
  if (update_queue != nsnull) {
    nsList *old_queue = update_queue;
    nsListItem *tmp_list = old_queue->getHead();

    update_queue = nsnull;

    if (tmp_list != nsnull) {
      while (tmp_list != nsnull)
      {
        nsWindow *window = (nsWindow *)tmp_list->getData();
        window->Update();
        window->mIsUpdating = PR_FALSE;

        tmp_list = tmp_list->getNext();
      }
    }
    delete old_queue;
  }
}

NS_IMETHODIMP nsWindow::GetAttention(void)
{

  GR_WINDOW_INFO wInfo;
  GrGetWindowInfo(mBaseWindow, &wInfo);
  while (wInfo.parent != GR_ROOT_WINDOW_ID)
    GrGetWindowInfo(wInfo.parent, &wInfo);
  GrRaiseWindow(wInfo.wid);
  return NS_OK;
}

void nsWindow::QueueDraw()
{
  if (!mIsUpdating) {
    if (update_queue == nsnull)
      update_queue = new nsList();
    update_queue->add((void *)this);
    mIsUpdating = PR_TRUE;
  }
}

void nsWindow::UnqueueDraw()
{
  if (mIsUpdating) {
    if (update_queue != nsnull)
      update_queue->remove((void *)this);
    mIsUpdating = PR_FALSE;
  }
}

///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////

nsWindow::nsWindow() : nsWidget()
{
  NS_INIT_REFCNT();
  mName.AssignWithConversion("nsWindow");
  mBackground = NS_RGB(255,255,255);

  mIsUpdating = PR_FALSE;
  mWindowType = eWindowType_child;
  mIsToplevel = PR_FALSE;
  mBlockFocusEvents = PR_FALSE;
}

nsWindow::~nsWindow()
{
  if (mIsUpdating)
    UnqueueDraw();
}

PRBool nsWindow::OnExpose(nsPaintEvent &event)
{
  return PR_TRUE;
}

NS_IMETHODIMP nsWindow::InvalidateRegion(const nsIRegion *aRegion,
                                         PRBool aIsSynchronous)
{
  mUpdateArea->Union(*aRegion);

  if (aIsSynchronous)
    Update();
  else
    QueueDraw();
  return NS_OK;
}

NS_IMETHODIMP nsWindow::SetFocus(PRBool aRaise)
{
  nsGUIEvent event;

  if (mFocusWindow == mBaseWindow)
    return NS_OK;

  if (mBlockFocusEvents == PR_TRUE)
    return NS_OK;

  mBlockFocusEvents = PR_TRUE;

  if (mFocusWindow) {
    nsWidget *oldFocusWidget = nsnull;

    oldFocusWidget = nsWidget::GetWidgetForWindow(mFocusWindow);

    if (oldFocusWidget) {
      event.message = NS_LOSTFOCUS;
      event.widget = oldFocusWidget;
      event.eventStructType = NS_GUI_EVENT;

      event.time = PR_Now();
      event.point.x = 0;
      event.point.y = 0;

      NS_ADDREF(oldFocusWidget);
      oldFocusWidget->DispatchWindowEvent(event);
      NS_RELEASE(oldFocusWidget);
      mFocusWindow = 0;
    }
  }

  if (mBaseWindow) {
    if (mFocusWindow != mBaseWindow) {
      mFocusWindow = mBaseWindow;
      GetAttention();
    }
  }

  event.message = NS_GOTFOCUS;
  event.widget = this;
  event.eventStructType = NS_GUI_EVENT;
  event.time = PR_Now();
  event.point.x = 0;
  event.point.y = 0;

  NS_ADDREF_THIS();
  DispatchWindowEvent(event);
  NS_RELEASE_THIS();

  event.message = NS_ACTIVATE;
  event.widget = this;
  event.eventStructType = NS_GUI_EVENT;
  event.time = PR_Now();
  event.point.x = 0;
  event.point.y = 0;

  NS_ADDREF_THIS();
  DispatchWindowEvent(event);
  NS_RELEASE_THIS();

  mBlockFocusEvents = PR_FALSE;

  return NS_OK;
}

void nsWindow::DestroyNativeChildren(void)
{
}

void nsWindow::DestroyNative(void)
{
  DestroyNativeChildren();

  if (mBaseWindow) {
    GrDestroyWindow(mBaseWindow);
    DeleteWindowCallback(mBaseWindow);
  }
}

NS_IMETHODIMP nsWindow::Resize(PRInt32 aWidth, PRInt32 aHeight, PRBool aRepaint)
{
  nsWidget::Resize(aWidth, aHeight, aRepaint);

  nsSizeEvent sevent;
  sevent.message = NS_SIZE;
  sevent.widget = this;
  sevent.eventStructType = NS_SIZE_EVENT;
  sevent.windowSize = new nsRect (0, 0, aWidth, aHeight);
  sevent.point.x = 0;
  sevent.point.y = 0;
  sevent.mWinWidth = aWidth;
  sevent.mWinHeight = aHeight;
  sevent.time = PR_Now();
  AddRef();
  OnResize(sevent);
  Release();
  delete sevent.windowSize;
  return NS_OK;
}

NS_IMETHODIMP nsWindow::Resize(PRInt32 aX,
                               PRInt32 aY,
                               PRInt32 aWidth,
                               PRInt32 aHeight,
                               PRBool   aRepaint)
{

  nsWidget::Resize(aX, aY, aWidth, aHeight, aRepaint);

  nsSizeEvent sevent;
  sevent.message = NS_SIZE;
  sevent.widget = this;
  sevent.eventStructType = NS_SIZE_EVENT;
  sevent.windowSize = new nsRect (aX, aY, aWidth, aHeight);
  sevent.point.x = aX;
  sevent.point.y = aY;
  sevent.mWinWidth = aWidth;
  sevent.mWinHeight = aHeight;
  sevent.time = PR_Now();
  AddRef();
  OnResize(sevent);
  Release();
  delete sevent.windowSize;
  return NS_OK;
}

NS_IMETHODIMP nsWindow::Invalidate(PRBool aIsSynchronous)
{
  mUpdateArea->SetTo(mBounds.x, mBounds.y, mBounds.width, mBounds.height);

  if (aIsSynchronous)
    Update();
  else
    QueueDraw();
  return NS_OK;
}

NS_IMETHODIMP nsWindow::Invalidate(const nsRect & aRect, PRBool aIsSynchronous)
{
  mUpdateArea->Union(aRect.x, aRect.y, aRect.width, aRect.height);
  if (aIsSynchronous)
    Update();
  else
    QueueDraw();

  return NS_OK;
}

NS_IMETHODIMP nsWindow::Update()
{
  if (mIsUpdating)
    UnqueueDraw();


  nsPaintEvent pevent;
  pevent.message = NS_PAINT;
  pevent.eventStructType = NS_PAINT_EVENT;
  pevent.rect = new nsRect (0,0,0,0);
  pevent.time = 0;
  pevent.region = mUpdateArea;

  if (!mUpdateArea->IsEmpty()) {
    PRUint32 numRects;
    mUpdateArea->GetNumRects(&numRects);

    if (numRects != 1 && numRects < 10) {
      nsRegionRectSet *regionRectSet = nsnull;

      if (NS_FAILED(mUpdateArea->GetRects(&regionRectSet))) {
        delete pevent.rect;
	return NS_ERROR_FAILURE;
      }

      PRUint32 len;
      PRUint32 i;

      len = regionRectSet->mRectsLen;

      for (i=0; i<len; ++i) {
	nsRegionRect *r = &(regionRectSet->mRects[i]);

	pevent.widget = this;
	pevent.rect->SetRect(r->x, r->y, r->width, r->height);

	pevent.time = PR_Now();
	AddRef();
	OnPaint(pevent);
	Release();
      }
      mUpdateArea->FreeRects(regionRectSet);
      mUpdateArea->SetTo(0,0,0,0);
      
      delete pevent.rect;
      return NS_OK;
    } else {
      PRInt32 x,y,w,h;
      mUpdateArea->GetBoundingBox(&x,&y,&w,&h);
      pevent.widget = this;
      pevent.rect->SetRect(x,y,w,h);
      AddRef();
      OnPaint(pevent);
      Release();
      mUpdateArea->SetTo(0,0,0,0);
    }
  } 
  
  delete pevent.rect;


  // The view manager also expects us to force our
  // children to update too!

  nsCOMPtr<nsIEnumerator> children ( getter_AddRefs(GetChildren()) );
  if (children)
  {
    children->First();
    do
    {
      nsISupports* child;
      if (NS_SUCCEEDED(children->CurrentItem(&child)))
      {
        nsWindow* childWindow = NS_STATIC_CAST(nsWindow*,child);
        NS_RELEASE(child);
        childWindow->Update();
      }
    } while (NS_SUCCEEDED(children->Next()));
  }
  return NS_OK;
}

NS_IMETHODIMP nsWindow::Scroll(PRInt32 aDx, PRInt32 aDy, nsRect *aClipRect)
{
  PRInt32 srcX, srcY, destX, destY, width, height;
  nsRect update, bound;
  
#ifdef DEBUG_mattw
  printf("\nnsWindow::Scroll(aDx=%d, aDy=%d)\n", aDx, aDy);
#endif

  if(aClipRect)
    bound.SetRect(aClipRect->x, aClipRect->y, aClipRect->width, aClipRect->height);
  else
    bound.SetRect(mBounds.x, mBounds.y, mBounds.width, mBounds.height);

#ifdef DEBUG_mattw
  printf("Scroll: Bounds: x=%d, y=%d, w=%d, h=%d\n",
  		bound.x, bound.y, bound.width, bound.height);
#endif

  if(mIsUpdating)
    UnqueueDraw();

  
  if (aDx < 0 || aDy < 0) {
    srcX = bound.x + ABS(aDx);
    srcY = bound.y + ABS(aDy);
    destX = bound.x;
    destY = bound.y;
    width = bound.width - ABS(aDx);
    height = bound.height - ABS(aDy);
  } else if (aDx > 0 || aDy > 0) {
    srcX = bound.x;
    srcY = bound.y;
    destX = bound.x + ABS(aDx);
    destY = bound.y + ABS(aDy);
    width = bound.width - ABS(aDx);
    height = bound.height - ABS(aDy);
  } else
    return NS_OK;

  // set update rect 
  if (aDx < 0) {
    update.SetRect(bound.width + aDx, 0, -aDx, bound.height);
  }
  else if (aDx > 0) {
    update.SetRect(0,0, aDx, bound.height);
  }

  if (aDy < 0) {
    update.SetRect(0, bound.height + aDy, bound.width, -aDy);
  }
  else if (aDy > 0) {
    update.SetRect(0,0, bound.width, aDy);
  }


//  static GR_SCREEN_INFO *sinfo = 0;
  static GR_GC_ID gc = 0;

  if(gc == 0)
  	gc = GrNewGC();
/*
  if(sinfo == 0) {
    sinfo = new GR_SCREEN_INFO;
    GrGetScreenInfo(sinfo);
  }
*/
  /**
   * We need to get the parent's parent to do dimension checks.
   * Another case that needs to be addressed is if the Main Window is off the
   * screen. If this is the case then we need to take it into account on top of
   * other factors. To capture all of these cases in one shot, I think it's
   easiest to check every window to the root window.
   */

#if 0
  GR_WINDOW_INFO winfo;
  GR_WINDOW_INFO pinfo;
  static GR_WINDOW_INFO tlinfo;
  static GR_WINDOW_ID topLevelID = 0;
  if(topLevelID == 0) {
    nsWidget *tlw = GetTopLevelWidget();
    NS_ASSERTION(tlw != 0, "Error getting top level widget");
    topLevelID = (GR_WINDOW_ID)tlw->GetNativeData(NS_NATIVE_WINDOW);
    GrGetWindowInfo(topLevelID, &tlinfo);
  }

  printf("Got top level Widget=%d\n", topLevelID);

  GrGetWindowInfo(mBaseWindow, &winfo);

  // Get out parent window info..
  GrGetWindowInfo(mParentWindow, &pinfo);	// my parent (scrollbar container)
  GrGetWindowInfo(pinfo.parent, &pinfo);	// my parent's parent 
  GrGetWindowInfo(pinfo.parent, &pinfo);	// my parent's parent's container

  int offTop = (winfo.y < 0 ? -winfo.y : 0); 
  int offLeft = (winfo.x < 0 ? -winfo.x : 0);
  int offBottom = (winfo.y + winfo.height > sinfo->rows 
  			? (winfo.y + winfo.height) - sinfo->rows : 0);
  int offRight = (winfo.x + winfo.width > sinfo->cols 
  			? (winfo.x + winfo.width) - sinfo->cols : 0);


  printf("Offscreen: top=%d, left=%d, bottom=%d, right=%d\n",
  		offTop, offLeft, offBottom, offRight);

  printf("Scroll: pre-mangled Blit: dx=%d, dy=%d, sx=%d, sy=%d, w=%d, h=%d\n",
  		destX, destY, srcX, srcY, width, height);

  // calc the blit coords
  if(offBottom > 0)
    height -= offBottom;
  if(offRight > 0) 
    width -= offRight;

  if(offLeft > 0) {
    if(aDx > 0) {
      destX += offLeft;
      srcX += offLeft;
    }
    if(aDx < 0) {
      destX += offLeft;
      srcX += offLeft;
    }
  }

  if(offTop > 0) {
    destY += offTop;
    srcY += offTop;
  }


#endif

#ifdef DEBUG_mattw
  printf("Scroll: Blit: dx=%d, dy=%d, sx=%d, sy=%d, w=%d, h=%d\n",
  		destX, destY, srcX, srcY, width, height);
#endif

  GrCopyArea(mBaseWindow, gc, destX, destY, width, height,
             mBaseWindow, srcX, srcY, MWROP_SRCCOPY);

#ifdef DEBUG_mattw
  printf("Scroll: Update: area: x=%d, y=%d, w=%d, h=%d\n",
  	update.x, update.y, update.width, update.height);
#endif

//  mUpdateArea->Offset(aDx, aDy);
  if(aDx > 0)
    Invalidate(bound, PR_TRUE);
  else
    Invalidate(update, PR_TRUE);

  //--------
  // Scroll the children
  nsCOMPtr<nsIEnumerator> children ( getter_AddRefs(GetChildren()) );
  if (children) {
    children->First();
    do {
      nsISupports* child;
      if (NS_SUCCEEDED(children->CurrentItem(&child))) {
        nsWindow* childWindow = NS_STATIC_CAST(nsWindow*,child);
        NS_RELEASE(child);
 
        nsRect bounds;
        childWindow->GetRequestedBounds(bounds);
        childWindow->Move(bounds.x + aDx, bounds.y + aDy);
        Invalidate(bounds, PR_TRUE);
      }
    } while (NS_SUCCEEDED(children->Next()));
  }

  return NS_OK;
}

NS_IMETHODIMP nsWindow::ScrollWidgets(PRInt32 aDx, PRInt32 aDy)
{
  return Scroll(aDx, aDy, nsnull);
}

NS_IMETHODIMP nsWindow::ScrollRect(nsRect &aSrcRect, PRInt32 aDx, PRInt32 aDy)
{
  return Scroll(aDx, aDy, &aSrcRect);
}

NS_IMETHODIMP nsWindow::SetTitle(const nsString &aTitle)
{
  if (!mBaseWindow)
    return NS_ERROR_FAILURE;

  GR_WM_PROPERTIES props;

  props.title = (GR_CHAR*)aTitle.ToNewCString();
  props.flags = GR_WM_FLAGS_TITLE;

  GrSetWMProperties(mBaseWindow, &props);

  return NS_OK;
}

NS_IMETHODIMP nsWindow::CaptureRollupEvents(nsIRollupListener *aListener,
                                            PRBool aDoCapture,
                                            PRBool aConsumeRollupEvent){
  if (aDoCapture) {
    gRollupListener = aListener;
    gRollupWidget = NS_STATIC_CAST(nsIWidget*,this);
    gRollupConsumeRollupEvent = PR_TRUE;
  }else{
    gRollupListener = nsnull;
    gRollupWidget = nsnull;
  }
  return NS_OK;
}

ChildWindow::ChildWindow() : nsWindow()
{
  mName.AssignWithConversion("nsChildWindow");
}
